import sys
from typing import NamedTuple

import pytest
from toy_crypto import nt, redundent


class TestFactor:
    def test_factor_small(self) -> None:
        vectors: list[tuple[int, list[tuple[int, int]]]] = [
            (2, [(2, 1)]),
            (10, [(2, 1), (5, 1)]),
            (9, [(3, 2)]),
            (1729, [(7, 1), (13, 1), (19, 1)]),
            (75, [(3, 1), (5, 2)]),
            (3557, [(3557, 1)]),
            (1999162, [(2, 1), (11, 3), (751, 1)]),
            ((2**32) - 1, [(3, 1), (5, 1), (17, 1), (257, 1), (65537, 1)]),
            (
                2**64 - 1,
                [
                    (3, 1),
                    (5, 1),
                    (17, 1),
                    (257, 1),
                    (641, 1),
                    (65537, 1),
                    (6700417, 1),
                ],
            ),
            (3 * 5 * 5 * 7, [(3, 1), (5, 2), (7, 1)]),
        ]

        for n, expected in vectors:
            assert nt.factor(n) == expected

    def test_OLF(self) -> None:
        vectors: list[tuple[int, int]] = [
            (22171, 1),
            (22171 * 45827 * 5483, 22171 * 5483),
        ]

        for n, expected in vectors:
            assert redundent.OLF(n) == expected

    def test_normalize(self) -> None:
        text_vectors: list[tuple[nt.FactorList, nt.FactorList]] = [
            (
                nt.FactorList([(2, 1), (11, 3), (751, 1)]),
                nt.FactorList([(2, 1), (11, 3), (751, 1)]),
            ),
            (
                nt.FactorList([(11, 3), (2, 1), (751, 1)]),
                nt.FactorList([(2, 1), (11, 3), (751, 1)]),
            ),
            (
                nt.FactorList([(11, 1), (2, 1), (11, 1), (751, 1), (11, 1)]),
                nt.FactorList([(2, 1), (11, 3), (751, 1)]),
            ),
        ]
        for tv in text_vectors:
            input, expected = tv
            assert input.normalize() == expected

    def test_add(self) -> None:
        # Mostly we are checking that result is normalized
        vectors = [
            (
                nt.FactorList([(11, 1), (2, 1), (11, 1)]),
                nt.FactorList([(751, 1), (11, 1)]),
                nt.FactorList([(2, 1), (11, 3), (751, 1)]),
            ),
            (
                nt.FactorList([(11, 1), (2, 1), (11, 1)]),
                nt.FactorList([]),
                nt.FactorList([(2, 1), (11, 2)]),
            ),
        ]
        for a, b, expected in vectors:
            assert a + b == expected
            assert b + a == expected

    def test_factor_large(self) -> None:
        """for input that should trigger Fermat's method,"""

        vectors: list[tuple[int, list[tuple[int, int]]]] = [
            (5483 * 5483, [(5483, 2)]),
            (5483 * 104243 * 11 * 11, [(11, 2), (5483, 1), (104243, 1)]),
            (2**32, [(2, 32)]),
            ((2**31) - 1, [(2147483647, 1)]),
            (104243 * 2147483647 * 2147483647, [(104243, 1), (2147483647, 2)]),
        ]
        for n, expected in vectors:
            assert nt.factor(n) == expected

    def test_phi(self) -> None:
        vectors = [
            ([(3, 1), (5, 1)], 8),
            ([(2, 2), (5, 1)], 8),
            ([(65537, 1)], 65536),
            ([(2, 1), (5, 1), (3, 1)], 8),
        ]
        for f, expected in vectors:
            assert nt.FactorList(f).phi == expected

    def test_n(self) -> None:
        vectors = [
            ([(3, 1), (5, 1)], 15),
            ([(2, 2), (5, 1)], 20),
            ([(65537, 1)], 65537),
        ]
        for input, expected in vectors:
            f = nt.FactorList(input)
            assert f.n == expected
            assert f.value() == expected

    def test_radical(self) -> None:
        vectors = [
            ([(3, 1), (5, 1)], [(3, 1), (5, 1)], 15),
            ([(2, 2), (5, 4)], [(2, 1), (5, 1)], 10),
            ([(2, 32)], [(2, 1)], 2),
        ]

        for input, exp_r, exp_rv in vectors:
            f = nt.FactorList(input)
            assert f.radical() == exp_r
            assert f.radical_value() == exp_rv

    def test_repr(self) -> None:
        vectors = [
            ([(5483, 2)], "5483^2"),
            ([(11, 2), (5483, 1), (104243, 1)], "11^2 * 5483 * 104243"),
            ([(2, 32)], "2^32"),
            ([(2, 6), (3, 3), (5, 3)], "2^6 * 3^3 * 5^3"),
        ]

        for input, expected in vectors:
            f = nt.FactorList(input)
            assert f.__repr__() == expected

    def test_coprimes(self) -> None:
        vectors = [
            (15, [1, 2, 4, 7, 8, 11, 13, 14]),
            (17, list(range(1, 17))),
            (30, [1, 7, 11, 13, 17, 19, 23, 29]),
            (2**5, list(range(1, 2**5, 2))),  # all odd numbers < 2^5
        ]

        for n, relprimes in vectors:
            f = nt.factor(n)
            assert list(f.coprimes()) == relprimes


class TestMath:
    def test_mr(self) -> None:
        vectors = [
            (512461, False),  # a Carmichael number
            (104297, True),
            (1180775137873020977354442912336269, True),
            (
                1025046047436407593990706183629376939000352221561805965005888683119,
                False,
            ),
        ]

        for n, is_prime in vectors:
            assert nt.isprime(n) == is_prime

    def test_gcd(self) -> None:
        vectors = [
            ((5, 50), 5),
            ((18, 12), 6),
            ((90, 409), 1),
            ((464, 403), 1),
            ((308, 463), 1),
            ((384, 320), 64),
            ((382, 472), 2),
        ]

        for input, expected in vectors:
            a, b = input
            assert nt.gcd(a, b) == expected

    def test_egcd(self) -> None:
        vectors = [
            (90, 409),
            (464, 403),
            (308, 463),
            (384, 320),
            (212, 351),
            (382, 472),
            (399, 107),
            (469, 269),
            (44, 163),
            (464, 194),
        ]

        for a, b in vectors:
            g, x, y = nt.egcd(a, b)
            assert g == nt.gcd(a, b)
            assert a * x + b * y == g

    def test_mod_sqrt(self) -> None:
        class TestVector(NamedTuple):
            a: int
            m: int
            expected: list[int]

        tests = [
            TestVector(58, 101, [82, 19]),
            TestVector(26, 101, []),
            TestVector(111, 113, [87, 26]),
            TestVector(55, 113, []),
            TestVector(30, 103, [37, 103 - 37]),
            TestVector(31, 103, []),
        ]

        for a, m, expected in tests:
            # first some verification of the roots in test vectors
            if expected:
                for root in expected:
                    assert a == root * root % m

            e: list[int] = [] if expected is None else sorted(list(expected))
            ret = nt.mod_sqrt(a, m)
            r: list[int] = [] if ret is None else sorted(list(ret))
            assert e == r

    def test_modinv(self) -> None:
        p = 868112830765445632873217196988651
        q = 1180775137873020977354442912336269

        p_inv = nt.modinv(p, q)
        q_inv = nt.modinv(q, p)

        assert (p_inv * p) % q == 1
        assert (q_inv * q) % p == 1

        assert p_inv > 0
        assert p_inv < q

        assert q_inv > 0
        assert q_inv < p

    def test_isqrt(self) -> None:
        p = 868112830765445632873217196988651
        vectors = [
            (2_000_000, 1414),
            (99, 9),  # check when n+1 is a a perfect square
            (100, 10),
            (1, 1),
            (0, 0),
            (p * p, p),
            ((p * p) + 1, p),
        ]

        for n, expected in vectors:
            assert nt.isqrt(n) == expected


class TestSieve:
    def test_sieve_30(self) -> None:
        s30 = nt.Sieve(30)
        expected = "001101010001010001010001000001"
        s30_count = 10

        assert s30.to01() == expected
        assert s30_count == s30.count

        # test that s30 still behaves as if it is 30 even if we create
        # a larger internal sieve
        s200 = nt.Sieve(200)

        # test that the larger sieve was indeed created
        s200_count = 46  # primes below 200
        assert s200.count == s200_count

        # large underlying sieve leaves s30's behavior unchanged
        assert s30.to01() == expected


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



from collections import namedtuple
from typing import Optional

from toy_crypto import rsa
from toy_crypto.nt import lcm, modinv
from toy_crypto.utils import Rsa129


class TestCitm:
    """Tests using Cat in the Middle story"""

    # Not really a great set of tests, since the data I am testing against
    # was created with some of the same code I'm testing

    e = 17

    class Critter:
        def __init__(
            self,
            factors: tuple[int, int],
            expected_N: int,
            expected_d: Optional[int] = None,
        ) -> None:
            self.factors = factors
            self.expected_N = expected_N
            self.expected_d = expected_d

            self.test_data: list[tuple[int, int]] = []

    e = 17

    patty = Critter((107, 151), 16157, 1403)
    patty.test_data = [(1234, 8900)]

    molly = Critter((97, 43), 4171, 593)
    molly.test_data = [(1313, 530), (1729, 2826)]

    mr_talk = Critter((47, 89), 4183, 1905)
    mr_talk.test_data = [(1729, 2016)]

    def test_encrypt(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)
            pubkey = key.pub_key

            for ptext, ctext in critter.test_data:
                assert ctext == pubkey.encrypt(ptext)

    def test_decrypt(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)

            for ptext, ctext in critter.test_data:
                assert ptext == key.decrypt(ctext)

    def test_N(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)
            pubkey = key.pub_key

            assert critter.expected_N == pubkey.N

    def test_d(self) -> None:
        for critter in [self.patty, self.molly, self.mr_talk]:
            p, q = critter.factors
            key = rsa.PrivateKey(p, q, pub_exponent=self.e)

            assert key._d == critter.expected_d


class TestSage:
    """Test data from SageMath Tutorial

    https://doc.sagemath.org/html/en/thematic_tutorials/numtheory_rsa.html

    The tutorial correctly points out that they way the primes
    were generated is inappropriate for real work.

    The tutorial uses phi directly instead of lcm(p-1, q-1).
    """

    # Don't use Mersenne primes in real life
    p = (2**31) - 1
    q = (2**61) - 1
    e = 1850567623300615966303954877
    m = 72697676798779827668  # message

    n = 4951760154835678088235319297
    phi = 4951760152529835076874141700
    d = 4460824882019967172592779313
    c = 630913632577520058415521090

    λ = lcm(p - 1, q - 1)

    def test_encrypt(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)
        pub_key = priv_key.pub_key

        assert pub_key.encrypt(self.m) == self.c

    def test_decrypt(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)

        assert priv_key.decrypt(self.c) == self.m

    def test_N(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)
        pub_key = priv_key.pub_key

        assert self.n == pub_key.N

    def test_d(self) -> None:
        priv_key = rsa.PrivateKey(self.p, self.q, self.e)

        # We (almost certainly) get a smaller d where the lcm check matters
        if self.phi == self.λ:
            assert priv_key._d == self.d


class TestMG1977:
    # encoder/decoder is in utils.Rsa129
    def test_magic(self) -> None:
        """Test the RSA-129 Challenge from Martin Gardner's 1977 article"""

        Challenge = namedtuple(
            "Challenge", ["modulus", "pub_exponent", "ctext"]
        )
        Solution = namedtuple("Solution", ["p", "q", "plaintext"])

        # From Martin Gardner's 1977
        challenge = Challenge(
            modulus=114381625757888867669235779976146612010218296721242362562561842935706935245733897830597123563958705058989075147599290026879543541,
            pub_exponent=9007,
            ctext=96869613754622061477140922254355882905759991124574319874695120930816298225145708356931476622883989628013391990551829945157815154,
        )

        # From Atkins et al 1995
        solution = Solution(
            p=3490529510847650949147849619903898133417764638493387843990820577,
            q=32769132993266709549961988190834461413177642967992942539798288533,
            plaintext="THE MAGIC WORDS ARE SQUEAMISH OSSIFRAGE",
        )

        pub_key = rsa.PublicKey(challenge.modulus, challenge.pub_exponent)

        # First test encryption
        plain_num: int = Rsa129.encode(solution.plaintext)
        ctext = pub_key.encrypt(plain_num)
        assert ctext == challenge.ctext

        # Now test decryption
        priv_key = rsa.PrivateKey(
            solution.p, solution.q, challenge.pub_exponent
        )
        assert priv_key == pub_key

        decrypted_num: int = priv_key.decrypt(ctext)
        decrypted: str = Rsa129.decode(decrypted_num)

        assert decrypted == solution.plaintext


class TestEq:
    # We will reuse magic values for this test, but anything should do
    p = 3490529510847650949147849619903898133417764638493387843990820577
    q = 32769132993266709549961988190834461413177642967992942539798288533
    e = 9007

    def test_pq_order(self) -> None:
        priv_pq = rsa.PrivateKey(self.p, self.q)
        priv_qp = rsa.PrivateKey(self.q, self.p)

        assert priv_pq == priv_qp

    def test_pub(self) -> None:
        priv_pq = rsa.PrivateKey(self.p, self.q, pub_exponent=self.e)
        priv_qp = rsa.PrivateKey(self.q, self.p, pub_exponent=self.e)

        assert priv_pq == priv_qp

    def test_pub_phi(self) -> None:
        """Key computed with 𝜑 is equivalent to key computed with λ."""

        key_lambda = rsa.PrivateKey(self.p, self.q, self.e)

        # can only construct key_phi by changing what would be private fields
        key_phi = rsa.PrivateKey(self.p, self.q, self.e)

        phi = (self.p - 1) * (self.q - 1)
        d_phi = modinv(key_phi.e, phi)

        key_phi._d = d_phi

        assert key_phi == key_lambda


class TestMisc:
    p = 3490529510847650949147849619903898133417764638493387843990820577
    q = 32769132993266709549961988190834461413177642967992942539798288533

    def test_default_e(self) -> None:
        default_e = rsa.default_e()
        assert default_e == 65537

        priv_key = rsa.PrivateKey(self.p, self.q)

        assert priv_key.e == default_e



import os
import sys
import timeit

PROJECT_PATH = os.getcwd()
SOURCE_PATH = os.path.join(PROJECT_PATH, "src")
sys.path.append(SOURCE_PATH)

from toy_crypto.nt import Sieve  # noqa: E402

setup = """
from toy_crypto.nt import Sieve
"""

FIRST_SIZE = 10_000_000
FINAL_SIZE = 100_000_000
TRIALS = 10

PRIME = 71
LEN = 100_000


def f1() -> None:
    Sieve.clear()
    Sieve(FIRST_SIZE)
    Sieve(FINAL_SIZE)


def f2() -> None:
    Sieve.clear()
    Sieve(FIRST_SIZE)
    Sieve.clear()
    Sieve(FINAL_SIZE)


def f_mod() -> int:
    r = LEN % PRIME
    if r == 0:
        return LEN
    return LEN + (PRIME - r)


t1 = timeit.Timer(stmt=f1).timeit(number=10)
print(f"time with cache: {t1}")

t2 = timeit.Timer(stmt=f2).timeit(number=10)
print(f"time with cache cleared: {t2}")



from toy_crypto.nt import lcm, modinv

_DEFAULT_E = 65537


def default_e() -> int:
    """Returns the default public exponent, 65537"""
    return _DEFAULT_E


class PublicKey:
    def __init__(self, modulus: int, public_exponent: int) -> None:
        """Public key from public values."""
        self._N = modulus
        self._e = public_exponent

    @property
    def N(self) -> int:
        """Public modulus N."""
        return self._N

    @property
    def e(self) -> int:
        """Public exponent e"""
        return self._e

    def encrypt(self, message: int) -> int:
        """Primitive encryption with neither padding nor nonce.

        :raises ValueError: if message < 0
        :raises ValueError: if message isn't less than the public modulus
        """

        if message < 0:
            raise ValueError("Positive messages only")
        if not message < self._N:
            raise ValueError("Message too big")

        return pow(base=message, exp=self._e, mod=self._N)

    def __eq__(self, other: object) -> bool:
        """True when each has the same modulus and public exponent.

        When comparing to a PrivateKey, this compares only the public parts.
        """
        if isinstance(other, PublicKey):
            return self.e == other.e and self.N == other.N

        return NotImplemented


class PrivateKey:
    def __init__(self, p: int, q: int, pub_exponent: int = _DEFAULT_E) -> None:
        """RSA private key from primes p and q.

        This does not perform any sanity checks on p and q.
        It is your responsibility to ensure that p and q are prime

        :raises ValueError: if e is not coprime with lcm(p - 1, q - 1).
        """

        self._p = p
        self._q = q
        self._e = pub_exponent

        self._N = self._p * self._q
        self._pubkey = PublicKey(self._N, self._e)

        self._dP = modinv(self._e, p - 1)
        self._dQ = modinv(self._e, (self._q - 1))
        self._qInv = modinv(self._q, self._p)

        try:
            self._d = self._compute_d()
        except ValueError:
            raise ValueError("p, q, and e are incompatible with each other ")

    @property
    def pub_key(self) -> PublicKey:
        """The public key corresponding to self.

        The public key does not contain any secrets.
        """

        return self._pubkey

    @property
    def e(self) -> int:
        """Public exponent."""
        return self._e

    def __eq__(self, other: object) -> bool:
        """True iff keys are mathematically equivalent

        Private keys with internal differences can behave identically
        with respect to input and output. This comparison will return
        True when they are equivalent in this respect.

        When compared to a PublicKey, this compares only the public part.
        """
        if isinstance(other, PrivateKey):
            return self.pub_key == other.pub_key

        if isinstance(other, PublicKey):
            return self.pub_key == other

        return NotImplemented

    def _compute_d(self) -> int:
        λ = lcm(self._p - 1, self._q - 1)
        try:
            return modinv(self.e, λ)
        except ValueError:
            raise ValueError("Inverse of e mod λ does not exist")

    def decrypt(self, ciphertext: int) -> int:
        """Primitive decryption."""

        if ciphertext < 1 or ciphertext >= self.pub_key.N:
            raise ValueError("ciphertext is out of range")

        # m =  pow(base=ciphertext, exp=self._d, mod=self._N)
        # but we will use the CRT
        # version comes from  rfc8017 §5.1.2

        m_1 = pow(ciphertext, self._dP, self._p)
        m_2 = pow(ciphertext, self._dQ, self._q)

        # I need to review CRT to see what this is for
        h = ((m_1 - m_2) * self._qInv) % self._p

        m = m_2 + self._q * h
        return m



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import tomllib

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../src"))

import toy_crypto  # noqa

from toy_crypto import __about__  # noqa: E402

version = __about__.__version__

# Pull general sphinx project info from pyproject.toml
# Modified from https://stackoverflow.com/a/75396624/1304076
with open("../../pyproject.toml", "rb") as f:
    toml = tomllib.load(f)

pyproject = toml["project"]

project = pyproject["name"]
release = version
author = ",".join([author["name"] for author in pyproject["authors"]])
copyright = f"2024 {author}"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions: list[str] = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
]

autodoc_typehints = "both"

extensions.append("sphinx.ext.intersphinx")
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "dns": ("https://dnspython.readthedocs.io/en/stable/", None),
}

rst_prolog = f"""
.. |project| replace:: **{project}**
.. |root| replace:: :mod:`toy_crypto`
.. _pyca: https://cryptography.io/en/latest/
.. _SageMath: https://www.sagemath.org
.. _primefac: https://pypi.org/project/primefac/
"""


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "nature"
html_static_path = ["_static"]



import sys

import pytest
from toy_crypto.ec import Curve
from toy_crypto.nt import Modulus


class TestSeriousCurve:
    """
    Tests involving example curve from Serious Cryptography
    """

    sc_parameters = (-4, 0, Modulus(191))
    sc_generator = (146, 131)
    curve = Curve(*sc_parameters)
    G = curve.point(*sc_generator)
    Px = 3
    Py = 46
    Qx = 146

    # Computed using SageMath
    sage_vectors: dict[int, tuple[int, int]] = {
        1: (146, 131),
        2: (8, 174),
        3: (137, 161),
        4: (4, 96),
        5: (61, 163),
        6: (34, 83),
        7: (171, 22),
        8: (18, 52),
        9: (174, 180),
        10: (77, 133),
        11: (186, 75),
        12: (136, 46),
        13: (29, 4),
        14: (48, 57),
        15: (155, 61),
        16: (158, 26),
        17: (71, 120),
        18: (102, 69),
        19: (179, 109),
        20: (15, 100),
        21: (122, 124),
        22: (3, 46),
        23: (87, 55),
        24: (79, 33),
        25: (74, 114),
        26: (10, 177),
        27: (53, 110),
        28: (32, 172),
        29: (56, 98),
        30: (25, 98),
        31: (99, 177),
        32: (100, 80),
        33: (42, 167),
        34: (52, 145),
        35: (165, 10),
        36: (156, 184),
        37: (110, 98),
        38: (160, 79),
        39: (82, 177),
        40: (98, 121),
        41: (19, 174),
        42: (50, 29),
        43: (164, 17),
        44: (6, 1),
        45: (111, 188),
        46: (67, 166),
        47: (140, 38),
        48: (2, 0),
        49: (140, 153),
        50: (67, 25),
        51: (111, 3),
        52: (6, 190),
        53: (164, 174),
        54: (50, 162),
        55: (19, 17),
        56: (98, 70),
        57: (82, 14),
        58: (160, 112),
        59: (110, 93),
        60: (156, 7),
        61: (165, 181),
        62: (52, 46),
        63: (42, 24),
        64: (100, 111),
        65: (99, 14),
        66: (25, 93),
        67: (56, 93),
        68: (32, 19),
        69: (53, 81),
        70: (10, 14),
        71: (74, 77),
        72: (79, 158),
        73: (87, 136),
        74: (3, 145),
        75: (122, 67),
        76: (15, 91),
        77: (179, 82),
        78: (102, 122),
        79: (71, 71),
        80: (158, 165),
        81: (155, 130),
        82: (48, 134),
        83: (29, 187),
        84: (136, 145),
        85: (186, 116),
        86: (77, 58),
        87: (174, 11),
        88: (18, 139),
        89: (171, 169),
        90: (34, 108),
        91: (61, 28),
        92: (4, 95),
        93: (137, 30),
        94: (8, 17),
        95: (146, 60),
    }

    def test_curve_repr(self) -> None:
        expected = "y^2 = x^3 - 4x + 0 (mod 191)"
        name = f"{self.curve}"
        assert name == expected

    def test_P_setup(self) -> None:
        P = self.curve.point(self.Px, self.Py)
        exp_P = 3, 46

        assert P.x == exp_P[0]
        assert P.y == exp_P[1]

    def test_compute_y(self) -> None:
        Qyy = self.curve.compute_y(self.Qx)
        assert Qyy is not None

        y0, y1 = Qyy

        Q0 = self.curve.point(self.Qx, y0)
        assert Q0.on_curve()

        Q1 = self.curve.point(self.Qx, y1)
        assert Q1.on_curve()

        assert (y0 + y1) % self.curve.p == 0

    def test_order(self) -> None:
        assert self.curve.order == 96

    def test_sums_on_curve(self) -> None:
        c = self.curve

        y = c.compute_y(self.Px)
        if not y:
            pytest.fail("failed to compute Py")
        else:
            Py = y[0]

        y = c.compute_y(self.Qx)
        if not y:
            pytest.fail("failed to compute Qy")
        else:
            Qy = y[0]

        P = c.point(self.Px, Py)
        Q = c.point(self.Qx, Qy)

        PpQ = P.add(Q)
        assert PpQ.on_curve()

        P2 = P.double()
        assert P2.on_curve()

    def test_generator(self) -> None:
        assert self.G.x == 146
        assert self.G.y == 131
        assert self.G.is_zero is False

        assert self.G.on_curve() is True

        for d in range(2, self.curve.order):
            dG = self.G.scaler_multiply(d)
            assert dG.on_curve()

    def test_double(self) -> None:
        G2 = self.G.double()
        assert (G2.x, G2.y) == (8, 174)

        G4 = G2.double()
        assert (G4.x, G4.y) == (4, 96)

    def test_doubling_only(self) -> None:
        doubled_only = [d for d in self.sage_vectors if d.bit_count() == 1]
        for d in doubled_only:
            x, y = self.sage_vectors[d]
            dG = self.G.scaler_multiply(d)
            assert dG.x == x
            assert dG.y == y

    def test_scaler_with_add(self) -> None:
        doubled_only = [d for d in self.sage_vectors if d.bit_count() != 1]
        for d in doubled_only:
            x, y = self.sage_vectors[d]
            dG = self.G.scaler_multiply(d)
            assert dG.x == x
            assert dG.y == y

    def test_pai(self) -> None:
        c = self.curve
        PaI = c.PAI
        P = c.point(self.Px, self.Py)
        negP = -P

        assert PaI.is_zero
        assert PaI + P == P
        assert P + PaI == P
        assert PaI == P - P
        assert negP + P == PaI

    def test_pai_immutable(self) -> None:
        c = self.curve
        Z = c.PAI
        P = c.point(self.Px, self.Py)
        with pytest.raises(NotImplementedError):
            Z += P

    def test_point_validation(self) -> None:
        with pytest.raises(ValueError):
            self.curve.point(self.Px, self.Py + 1)


def test_curve_validation() -> None:
    with pytest.raises(ValueError):
        Curve(2, 3, p=Modulus(31 * 73))

    with pytest.raises(ValueError):
        Curve(2, 3, p=Modulus(5))


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



"""Utility functions"""

from collections.abc import Iterator
import itertools
from typing import Optional, Self
from toy_crypto.types import Byte


def lsb_to_msb(n: int) -> Iterator[int]:
    """
    0s and 1s representing bits of n, starting with the least significant bit.

    :raises TypeError: if n is not an integer.
    :raises ValueError: if n is negative.
    """
    if not isinstance(n, int):
        raise TypeError("n must be an integer")
    if n < 0:
        raise ValueError("n cannot be negative")

    while n > 0:
        yield n & 1
        n >>= 1


def digit_count(n: int, base: int = 10) -> int:
    """returns the number of digits (base b) of integer n.

    :raises ValueError: if base < 2
    :raises TypeError: if n is not an integer

    """

    if base < 2:
        raise ValueError("base must be at least 2")

    if not isinstance(n, int):
        raise TypeError("n must be an integer")

    n = abs(n)

    # Handling this case separately seems better than trying
    # to simulate a do-while in Python.
    if n == 0:
        return 1

    digits = 0
    while n > 0:
        digits += 1
        n = n // base
    return digits


class Xor:
    """Iterator that spits out xor of message with (repeated) pad.

    The iterator will run through successful bytes of message
    xor-ing those with successive bytes of pad, repeating
    pad if pad is shorter than message.

    Each iteration returns a non-negative int less than 256.
    """

    def __init__(
        self,
        message: Iterator[Byte] | bytes,
        pad: bytes,
    ) -> None:
        # Convert message to Iterator if needed
        self._message: Iterator[Byte] = iter(message)
        self._pad: Iterator[Byte] = itertools.cycle(pad)

    def __next__(self) -> Byte:
        b, p = next(zip(self._message, self._pad))
        return Byte(b ^ p)

    def __iter__(self: Self) -> Self:
        return self


def xor(message: bytes | Iterator[Byte], pad: bytes) -> bytes:
    """Returns the xor of message with a (repeated) pad.

    The pad is repeated if it is shorter than m.
    This can be thought of as bytewise Vigenère.
    """
    return bytes([b for b in Xor(message, pad)])


def hamming_distance(a: bytes, b: bytes) -> int:
    """Hamming distance between byte sequences of equal length.

    :raises ValueError: if len(a) != len(b).
    """

    if len(a) != len(b):
        raise ValueError("Lengths are unequal")

    # hamming distance will be the number of 1 bits in a xor b
    db: bytes = xor(a, b)
    # bit_count is only defined for ints, so
    return int.from_bytes(db, signed=False).bit_count()


class Rsa129:
    """Text encoder/decoder used in RSA-129 challenge.

    Encoding scheme from Martin Gardner's 1977 article.
    """

    _abc: list[str] = list(" ABCDEFGHIJKLMNOPQRSTUVWXYZ")
    _abc_index: Optional[dict[str, int]] = None

    @classmethod
    def _make_index(cls) -> dict[str, int]:
        if cls._abc_index is None:
            cls._abc_index = {c: i for i, c in enumerate(cls._abc)}
        return cls._abc_index

    def __init__(self) -> None:
        self._make_index()

    @classmethod
    def encode(cls, text: str) -> int:
        """Encode text"""

        indx = cls._make_index()
        result = 0
        for c in text:
            result *= 100
            result += indx[c]
        return result

    @classmethod
    def decode(cls, number: int) -> str:
        """Decode text."""
        chars: list[str] = []
        while True:
            number, rem = divmod(number, 100)
            chars.append(cls._abc[rem])
            if number == 0:
                break
        return "".join(reversed(chars))
